import os, re
from agent1c_metrics import read1c, settings
from pathlib import Path
from datetime import datetime, timedelta, time

def parse_1CV8Clst(cfg_file):
    cfgdata = {'cluster':[],'bases':[]}

    if not os.path.isfile(cfg_file):
        return cfgdata | {'message':f'File not exists: {cfg_file}'}
    
    try:
        data = read1c.lts(cfg_file)
    except read1c.ParseJSONException as e:
        return cfgdata | {'error':'Error when reading 1c file','file':cfg_file,'content':e.original_text}

    if data:

        cl_data = data[1]
    
        cluster_fields_list = ['id','name','port','host','p4','secure_connection','workers_restart_period','problem_workers_kill_after','unknown_8','unknown_9','unknown_10','cluster_managers','load_sharing_mode','unknown_13','force_kill_problem_workers','write_process_dump_due_to_memory_overload']
        cluster_fields = dict(map(lambda i: (i,lambda x: x),cluster_fields_list))
        # customization
        cluster_fields['host'] = lambda x: x.lower()

        cfgdata['cluster'] = {list(cluster_fields)[i] if i<len(cluster_fields) else f'p{i}':list(cluster_fields.values())[i](cl_data[i]) if i<len(cluster_fields) else cl_data[i] for i in range(len(cl_data))}

        ib_fields_list = ['id','name','discription','dbtype','dbserver','dbname','dbuser','dbpasshash','dbstr','unknown_p1','block','block_tasks','share_licence_by_server','unknown_p3','use_external_management','security_profile','security_profile_for_secure_mode','settings_version','reserving_workers']
        ib_block_fields = {
            'block_sessions':lambda x: bool(x),
            'block_start': lambda x: datetime.fromisoformat(str(x)[:4]+'-'+str(x)[4:6]+'-'+str(x)[6:8]+'T'+str(x)[8:10]+':'+str(x)[10:12]+':'+str(x)[12:]),
            'block_end': lambda x: datetime.fromisoformat(str(x)[:4]+'-'+str(x)[4:6]+'-'+str(x)[6:8]+'T'+str(x)[8:10]+':'+str(x)[10:12]+':'+str(x)[12:]),
            'block_message': lambda x: x,
            'block_unlock_code': lambda x: x,
            'block_param': lambda x: x,
        }
        ib_fields = dict(map(lambda i: (i,lambda x: x),ib_fields_list))
        # customization
        ib_fields['dbserver'] = lambda x: x.lower()
        ib_fields['block'] = lambda x: {list(ib_block_fields)[i] if i<len(ib_block_fields) else f'block_p{i}':list(ib_block_fields.values())[i](x[i]) if i<len(ib_block_fields) else x[i] for i in range(len(x))}

        for ibdata in data[2][1:]:
            ib = {list(ib_fields)[i] if i<len(ib_fields) else f'p{i}':list(ib_fields.values())[i](ibdata[i]) if i<len(ib_fields) else ibdata[i] for i in range(len(ibdata))}
            cfgdata['bases'].append(ib)

    return cfgdata

def parse_1cv8wsrv(lst_file):
    lst_data = {'clusters':[]}

    if not os.path.isfile(lst_file):
        return lst_data | {'message':f'File not exists: {lst_file}'}
    
    try:
        data = read1c.lts(lst_file)
    except read1c.ParseJSONException as e:
        return lst_data | {'error':'Error when reading 1c file','file':lst_file,'content':e.original_text}

    cluster_fields_list = ['id','name','port','host','p4','secure_connection','workers_restart_period','problem_workers_kill_after','unknown_8','unknown_9','unknown_10','cluster_managers','load_sharing_mode','unknown_13','force_kill_problem_workers','write_process_dump_due_to_memory_overload']
    cluster_fields = dict(map(lambda i: (i,lambda x: x),cluster_fields_list))
    # customization
    cluster_fields['host'] = lambda x: x.lower()

    for cl_data in data[0][1:]:
        lst_data['clusters'].append({list(cluster_fields)[i] if i<len(cluster_fields) else f'p{i}':list(cluster_fields.values())[i](cl_data[i]) if i<len(cluster_fields) else cl_data[i] for i in range(len(cl_data))})
    
    return lst_data

def get_data():

    data = {'clusters':[]}

    for path1c in settings['folders']:
        
        filepath_1cv8wsrv = os.path.join(path1c,'1cv8wsrv.lst')
        
        if not os.path.isfile(filepath_1cv8wsrv): # we are in reg_1541 or similar folder
            path1c = Path(path1c).parent.absolute()
            filepath_1cv8wsrv = os.path.join(path1c,'1cv8wsrv.lst')
        
        cluster_info = parse_1cv8wsrv(filepath_1cv8wsrv)
        
        for cluster_item in cluster_info['clusters']:
            filepath_1CV8Clst = os.path.join(path1c,f"reg_{cluster_item['port']}",'1CV8Clst.lst')
            cluster_item['data'] = parse_1CV8Clst(filepath_1CV8Clst)

            # add info of LOG size and type
            for ib in cluster_item['data']['bases']:
                ibpath = os.path.join(path1c,f"reg_{cluster_item['port']}",ib['id'],'1Cv8Log')

                # get log type
                ib['logtype'] = 'txt' if os.path.isfile(os.path.join(ibpath,'1Cv8.lgf')) else 'sqlite'
                
                # get size
                ib['logsize'] = 0
                ib['mtime'] = 0
                if os.path.exists(ibpath):
                    for ele in os.scandir(ibpath):
                        #print('-',ele)
                        ib['logsize'] += os.path.getsize(ele)
                        ib['mtime'] = max(ib['mtime'],os.path.getmtime(ele))

                ib['mtime_iso'] = datetime.fromtimestamp(ib['mtime'])
                ib['mtime_lastday'] = ib['mtime_iso'] > datetime.today() - timedelta(days=1)
                ib['mtime_lastweek'] = ib['mtime_iso'] > datetime.today() - timedelta(days=7)
                ib['mtime_lastmonth'] = ib['mtime_iso'] > datetime.today() - timedelta(days=31)
                ib['inactivity'] = (datetime.now() - ib['mtime_iso']).days

        data['clusters'] += cluster_info['clusters']

    return data



from importlib import resources
from yaml import load, dump
import os

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

__settings_default = load(resources.read_text("agent1c_metrics", "config.yaml"),Loader=Loader)
__settings_filename = "agent1c_metrics_config.yaml"

# create settings file, if not exists
if os.path.isfile(__settings_filename):
    with open(__settings_filename,"r") as settings_file:
        settings = load(settings_file.read(),Loader=Loader)
else:
    with open(__settings_filename,"w") as settings_file:
        settings = __settings_default
        settings_file.write(dump(settings))

# Version of the package
__version__ = "0.4.2"



from fastapi import APIRouter
from agent1c_metrics import settings, __settings_filename
from pydantic import BaseModel

class Settings(BaseModel):
    folders: list[str]

router = APIRouter(prefix="/settings", tags=["settings"])

@router.get('/', response_model=Settings)
def get_settings():
    return settings

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

@router.post('/', response_model=Settings)
def add_settings(settings_data: Settings):
    __add_settings = dict(settings_data)
    for k in __add_settings:
        settings[k] += __add_settings[k]
        settings[k] = list(dict.fromkeys(settings[k])) # unique

    with open(__settings_filename,"w") as settings_file:
        settings_file.write(dump(settings,Dumper=Dumper))

    return settings

@router.delete('/', response_model=Settings)
def delete_settings(settings_data: Settings):
    __del_settings = dict(settings_data)
    for k in __del_settings:
        settings[k] = [item for item in settings[k] if item not in __del_settings[k]]

    with open(__settings_filename,"w") as settings_file:
        settings_file.write(dump(settings,Dumper=Dumper))

    return settings


import win32serviceutil
import win32service
import win32event
import servicemanager
import socket

import uvicorn

class AppServerSvc (win32serviceutil.ServiceFramework):
    _svc_name_ = "agent1c_metrics"
    _svc_display_name_ = "Agent 1C metrics service"
    _svc_port = 8144
    _svc_host = '0.0.0.0'
    _svc_reload = False

    def __init__(self,args):
        win32serviceutil.ServiceFramework.__init__(self,args)
        self.hWaitStop = win32event.CreateEvent(None,0,0,None)
        socket.setdefaulttimeout(60)

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                              servicemanager.PYS_SERVICE_STARTED,
                              (self._svc_name_,''))
        self.main()

    def main(self):
        uvicorn.run("agent1c_metrics.viewer:app", port=self._svc_port, log_level="info", host=self._svc_host, reload=self._svc_reload)

def svc() -> None:
    """
    Manage windows service
    """
    AppServerSvc
    win32serviceutil.HandleCommandLine(AppServerSvc)

if __name__ == "__main__":
    svc()


import json, re

class ParseJSONException(json.JSONDecodeError):
    def __init__(self, original_text, message="Unable to parse 1c data"):
        self.original_text = original_text
        self.message = message
        super().__init__(self.message)

def lts(lts_file):

    str_original = ''
    with open(lts_file,encoding='utf-8') as cfg:
        str_original = cfg.read().replace("\r\n",'\n').replace('\n','').replace('\uFEFF','').replace('{','[').replace('}',']')
    str_original = re.sub(r"([0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12})",wrap_by_quotes,str_original)
    str_original = re.sub(r"[^,\[](\"\")",replace_double_quotes,str_original)
    str_original = re.sub(r",(0\d+)",wrap_by_quotes,str_original)
    str_original = re.sub(r"(\\)[^\"]",wrap_backslash,str_original)
    
    try:
        data = json.loads(str_original)
    except:
        raise ParseJSONException(str_original)

    return data

def wrap_by_quotes(match_obj):
    environment = match_obj.group()
    value = match_obj.group(1)
    replacement = environment.replace(value,f'"{value}"')
    if value is not None:
        return replacement

def wrap_backslash(match_obj):
    environment = match_obj.group()
    value = match_obj.group(1)
    replacement = environment.replace(value,f'\\\\')
    if value is not None:
        return replacement

def replace_double_quotes(match_obj):
    value = match_obj.group()
    if value is not None:
        return value.replace('""','\\"')


from fastapi import APIRouter
from fastapi.responses import PlainTextResponse
from string import Template
from agent1c_metrics.reader import get_data

router = APIRouter(prefix="/metrics", tags=["metrics"])

@router.get('/', response_class=PlainTextResponse)
async def metrics():
    result = [
        "# HELP agent1c_metrics_infobase_logsize Cumulative size of 1cv8log folder.",
        "# TYPE agent1c_metrics_infobase_logsize gauge",
        
        "# HELP agent1c_metrics_infobase_changed_lastday A metric with a boolean value labeled with ibase information.",
        "# TYPE agent1c_metrics_infobase_changed_lastday gauge",
        
        "# HELP agent1c_metrics_infobase_changed_lastweek A metric with a boolean value labeled with ibase information.",
        "# TYPE agent1c_metrics_infobase_changed_lastweek gauge",
        
        "# HELP agent1c_metrics_infobase_changed_lastmonth A metric with a boolean value labeled with ibase information.",
        "# TYPE agent1c_metrics_infobase_changed_lastmonth gauge",

        "# HELP agent1c_metrics_infobase_block_sessions A metric with a boolean value labeled with ibase information.",
        "# TYPE agent1c_metrics_infobase_block_sessions gauge",

        "# HELP agent1c_metrics_infobase_block_tasks A metric with a boolean value labeled with ibase information.",
        "# TYPE agent1c_metrics_infobase_block_tasks gauge",

        "# HELP agent1c_metrics_infobase_inactivity A metric with a int value contains days of inactivity of the base calculated from log modification time.",
        "# TYPE agent1c_metrics_infobase_inactivity gauge",
    ]

    #result.append("# TYPE logtype_by_infobase gauge")
    templates = [
        Template("agent1c_metrics_infobase_logsize{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $logsize"),
        Template("agent1c_metrics_infobase_changed_lastday{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $mtime_lastday"),
        Template("agent1c_metrics_infobase_changed_lastweek{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $mtime_lastweek"),
        Template("agent1c_metrics_infobase_changed_lastmonth{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $mtime_lastmonth"),
        Template("agent1c_metrics_infobase_block_sessions{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $block_sessions"),
        Template("agent1c_metrics_infobase_block_tasks{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $block_tasks"),
        Template("agent1c_metrics_infobase_inactivity{host=\"$host\",port=\"$port\",ibname=\"$name\",ibid=\"$id\",type=\"$logtype\",dbserver=\"$dbserver\",dbname=\"$dbname\"} $inactivity"),
    ]
    #t_logtype = Template("agent1c_metrics_logtype_by_infobase{host=\"$host\",port=\"$port\",ibname=\"$ibname\",ibid=\"$ibid\",type=\"$type\"} 1")
    data = get_data()
    for cluster_info in data['clusters']:
        for ib in cluster_info['data']['bases']:
            for t in templates:
                cluster_prepared_info = { key:val for key, val in cluster_info['data']['cluster'].items() if key not in ['id','name'] }
                ib_prepared_data = dict(map(lambda k: (k,(1 if ib[k] else 0) if type(ib[k])==bool else ib[k]),ib))
                ib_prepared_block_data = dict(map(lambda k: (k,(1 if ib_prepared_data['block'][k] else 0) if type(ib_prepared_data['block'][k])==bool else ib_prepared_data['block'][k]),ib_prepared_data['block']))
                result.append(t.safe_substitute(**cluster_prepared_info,**ib_prepared_data,**ib_prepared_block_data))
    return PlainTextResponse('\n'.join(result))



from fastapi import FastAPI
from agent1c_metrics.api import settings_api, metrics
from agent1c_metrics import __version__, settings

app = FastAPI(title="1C-agent metrics")
app.include_router(settings_api.router)
app.include_router(metrics.router)
app.version = __version__

from agent1c_metrics.reader import get_data
from agent1c_metrics import settings

@app.get("/")
async def root():

    result = get_data()
    
    return result

if __name__ == "__main__":
    print('Run __main__.py module instead.')
else:
    print(settings)


import click, uvicorn

@click.command()
@click.option("--reload", is_flag=True, help="Reload if code changes")
@click.option("--host", default='0.0.0.0', type=str, required=False, help="Host for reading")
@click.option("--port", default='8144', type=str, required=False, help="Port for reading")
def main(reload:bool,host:str, port:str) -> None:
    uvicorn.run("agent1c_metrics.viewer:app", port=8144, log_level="info", host=host, reload=reload)

if __name__ == "__main__":
    main()